from __future__ import print_function, unicode_literals

import six
from django.core.management.base import BaseCommand

from wrapper_tag import utils


class Command(BaseCommand):
    help = "Show documentation to tags"

    def add_arguments(self, parser):
        # Positional arguments
        parser.add_argument('tags', nargs='*')

    def handle(self, *args, **options):
        libs = utils.get_config().get_templatetag_libs()

        for name, lib in six.iteritems(libs):
            for tag_name, compilation_func in six.iteritems(lib.tags):
                if tag_name in options['tags']:
                    print("\n{}".format(tag_name))
                    print("-" * len(tag_name))
                    print("To use this tag, put {{% load {} %}} in your template before using the tag.\n".format(name))
                    print(compilation_func.__doc__)



"""
Documentation generator for WrapperTag
"""
import textwrap
from collections import OrderedDict
from operator import attrgetter

import six
from django.utils.encoding import force_text, smart_text
from django.utils.html import format_html
from django.utils.safestring import mark_safe


class ArgumentsGroup(object):
    """
    ArgumentGroup groups classes of arguments in documentation


    Every custom argument class needs to provide `doc_group` attribute, so docgen knows how to group arguments in
    documentation.

    Example::

        class Event(Argument):
            doc_group = docgen.ArgumentsGroup(_('Events'), help_text=_('Event handlers that are available on rendered '
                                                               'tag `events` dictionary'), priority=0)

    This will ensure that all events are groupped in group 'Events' in tag documentation.
    """

    help_text = None
    priority = None
    title = None

    def __init__(self, title, help_text=None, priority=100):
        """

        :param title: title of arguments group
        :param help_text: short description
        :param priority: priority how to order groups
        """
        self.title = title
        self.help_text = help_text
        self.priority = priority

    def render(self, arguments):
        """
        render renders documentation for arguments
        :param arguments:
        :return:
        """
        doc = format_html("**{}**", self.title) + "\n\n"

        if self.help_text:
            doc += "*{}*\n\n".format(force_text(self.help_text))

        for argument in arguments:
            doc += "* {doc}\n".format(doc=smart_text(argument.gen_doc()))

        return mark_safe(doc)

    @classmethod
    def group_arguments(cls, tag_cls):
        """
        Group tag arguments by `doc_group` and return ordered dict where key is ArgumentsGroup instance
        :param tag_cls:
        :return:
        """
        groups = {}
        for _, argument in six.iteritems(tag_cls.arguments):
            if groups.get(argument.doc_group, None) is None:
                groups[argument.doc_group] = []

            groups[argument.doc_group].append(argument)

        result = OrderedDict()
        for key in sorted(groups.keys(), key=attrgetter('priority'), reverse=True):
            result[key] = sorted(groups[key], key=attrgetter('name'))

        return result


def generate(tag_cls):
    """
    generate generates documentation for given wrapper tag class
    :param tag_cls: wrapper_tag class
    :return:
    """

    doc = textwrap.dedent(tag_cls.__doc__ or '').strip()

    arguments_doc = ""

    for ag, arguments in six.iteritems(ArgumentsGroup.group_arguments(tag_cls)):
        filtered = filter(lambda x: not x.readonly, arguments)
        if not filtered:
            continue
        rendered_group = ag.render(filtered)
        arguments_doc += rendered_group + "\n"

    arguments_doc = arguments_doc.strip()

    # if {argumets} found in documentation it will be replaced with arguments documentation, otherwose it will
    # be appended to documentation
    if '{arguments}' in doc:
        doc = doc.replace('{arguments}', arguments_doc)
    else:
        doc = doc + "\n\n" + arguments_doc

    return doc



"""
Various helper utilities for wrapper tag
"""
from __future__ import absolute_import, print_function, unicode_literals
import re

from django import get_version
from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.template import Context, RequestContext, Template
from django.template import TemplateSyntaxError
from django.template.base import token_kwargs
from django.template.context import BaseContext
from django.template.loader import get_template

from distutils.version import LooseVersion

from django.utils import inspect
from django.utils import six
from django.utils.functional import cached_property

from logging import getLogger


DJANGO_1_9_OR_HIGHER = (
    LooseVersion(get_version()) >= LooseVersion('1.9')
)

# pattern for camel_to_separated
CAMEL_TO_SEPARATED = re.compile('((?<=[a-z0-9])[A-Z]|(?!^)[A-Z](?=[a-z]))')

# regex type for isinstance
REGEX_TYPE = type(re.compile(''))


class NULL:
    """
    Internal type
    """


def camel_to_separated(s, separator='.'):
    """
    Returns dotted_case from CamelCase
    :param separator: separator to join multiple parts of name
    :param s: string to convert
    :return:
    """
    return CAMEL_TO_SEPARATED.sub(separator + r'\1', s).lower()


def flatten_context(context):
    def do_flatten(context):
        flat = {}
        for d in context.dicts:
            if isinstance(d, (Context, RequestContext)):
                flat.update(do_flatten(d))
            else:
                flat.update(d)
        return flat

    if callable(getattr(context, 'flatten', None)) and DJANGO_1_9_OR_HIGHER:
        return context.flatten()
    elif isinstance(context, BaseContext):
        return do_flatten(context)
    return context


def get_config():
    """
    Return wrapper tag config
    :return wrapper_tag.apps.WrapperTagConfig:
    """
    return apps.get_app_config('wrapper_tag')


def parse_bits(parser, bits, name):
    """
    Taken from django code and enhanced (params accept also wildcard on the end of string

    Parses bits for template tag helpers (simple_tag, include_tag and
    assignment_tag), in particular by detecting syntax errors and by
    extracting positional and keyword arguments.
    :param name:
    :param bits:
    :param parser:
    """
    args = []
    kwargs = {}
    for bit in bits:
        # First we try to extract a potential kwarg from the bit
        kwarg = token_kwargs([bit], parser)
        if kwarg:
            # The kwarg was successfully extracted
            param, value = list(six.iteritems(kwarg))[0]
            if param in kwargs:
                # The keyword argument has already been supplied once
                raise TemplateSyntaxError(
                    "'%s' received multiple values for keyword argument '%s'" %
                    (name, param))
            else:
                # All good, record the keyword argument
                kwargs[str(param)] = value
        else:
            if kwargs:
                raise TemplateSyntaxError(
                    "'%s' received some positional argument(s) after some "
                    "keyword argument(s)" % name)
            else:
                # Record the positional argument
                args.append(parser.compile_filter(bit))

    return args, kwargs


class TemplateMixin(object):

    template_errors = {
        'no_template': 'Please provide template or template_name',
    }

    def __init__(self, template=None, template_name=None):
        self._template = template
        self._template_name = template_name

    @cached_property
    def template(self):
        if self._template:
            return Template(self._template)
        elif self._template_name:
            return get_template(self._template_name)


def get_sub_logger(logger, *names):
    """
    Return sub logger
    :param logger:
    :param name:
    :return:
    """
    if not names:
        return logger
    return getLogger('{}.{}'.format(logger.name, '.'.join(names)))


def is_seq(iterable):
    """
    Returns whether iterable is list|tuple
    :param iterable:
    :return:
    """
    return isinstance(iterable, (list, tuple))


def is_template_debug():
    """
    Returns whether template debug is enabled
    :return:
    """
    return bool(getattr(settings, 'TEMPLATE_DEBUG', False))


def is_rendered_tag(value, tag):
    """
    Returns whether it was stored with "as" attribute
    :param value:
    :param tag:
    :return:
    """
    from wrapper_tag import RenderedTag
    if not isinstance(value, RenderedTag):
        return False

    return tag == value.tag


def verify_func_signature(func, *args, **kwargs):
    """
    verify signature of function
    :param func:
    :return:
    """
    spec = inspect.getargspec(func)
    need_star_kwargs = kwargs.pop('need_star_kwargs', False)
    exact_names = kwargs.pop('exact_names', False)
    prefix = kwargs.pop('prefix', '')
    enable_if = kwargs.pop('enable_if', True)
    enabled = enable_if() if callable(enable_if) else enable_if

    if not enabled:
        return

    if spec.args and spec.args[0] == 'self':
        spec_args = spec.args[1:]
    else:
        spec_args = spec.args

    func_name = func.__name__
    if need_star_kwargs and not spec.keywords:
        raise ImproperlyConfigured('{}function {}, does not accept **kwargs.'.format(prefix, func_name))

    if len(spec_args) != len(args):
        raise ImproperlyConfigured('{}function {}, signature does not match to: {}.'.format(prefix, func_name, args))

    if exact_names:
        for i in range(len(args)):
            if spec_args[i] != args[i]:
                raise ImproperlyConfigured('{}function {}, must have exact arg names: {}, got: {}.'.format(
                    prefix, func_name, args, spec_args))


def find_elements(item, iterable):
    """
    Finds elements with wildcard in iterable (list, tuple..)
    Both item and iterable can be wildcarded.

    :param item: item to be found
    :param iterable: list/tuple of elements to find in..
    :return: list of found
    """
    result = []

    if not isinstance(item, REGEX_TYPE):
        pattern = '^{}$'.format(item.replace('+', '.+').replace('*', '.*'))
        item = re.compile(pattern)

    for value in iterable:
        if item.match(value):
            result.append(value)

    return result



from __future__ import unicode_literals, print_function, absolute_import
import pkg_resources  # part of setuptools
from wrapper_tag.arguments import Keyword, KeywordGroup, Event, Method
from wrapper_tag.rendered import RenderedTag
from wrapper_tag.tag import Tag
from wrapper_tag.utils import parse_bits, is_seq, is_template_debug
from logging import getLogger
from django.core.exceptions import ImproperlyConfigured

# read version from setuptools
__version__ = pkg_resources.require("django_wrapper_tag")[0].version

default_app_config = 'wrapper_tag.apps.WrapperTagConfig'

__all__ = (
    'Keyword', 'KeywordGroup', 'Event', 'Method',
    'RenderedTag',
    'Tag',
    'register_tag',
)

logger = getLogger("wrapper_tag")


def register_tag(library, **kwargs):
    """
    Register WrapperTag class
    :param library:
    :param kwargs:
    :return:

    @TODO: update docs for given wrapper tag
    """

    def get_library_name(cls):
        """
        Returns library name
        :param cls:
        :return:
        """
        return repr(cls).split('.')[-2]

    def wrap(cls):
        if not issubclass(cls, Tag) and is_template_debug():
            raise ImproperlyConfigured("Class {} is not WrapperTag instance".format(cls))

        from wrapper_tag import docgen

        cls.__doc__ = docgen.generate(cls)

        # add main library tag
        library.tag(cls.options.start_tag, cls)

        logger.debug("Registered `%s`=>`%s` wrapper tag.", cls.options.start_tag, cls.options.end_tag)

        # # set library_name
        # cls.options.library_name = get_library_name(cls)

        # aliases = []
        # for alias in cls.options.aliases:
        #     if is_seq(alias):
        #         start_tag, end_tag = alias[0], alias[1]
        #     else:
        #         start_tag, end_tag = alias, 'end:{}'.format(alias)
        #     aliases.append({
        #         'start_tag': start_tag,
        #         'end_tag': end_tag,
        #     })
        #
        # doc = cls.gen_doc(aliases=aliases)

        # # register_tag also aliases (if available)
        # for alias in aliases:
        #     new_wrapper_tag_class = type(str('Alias_{}_{}'.format(alias, cls.__name__)), cls.__bases__,
        #                                  dict(cls.__dict__))
        #     new_wrapper_tag_class.options.start_tag = alias['start_tag']
        #     new_wrapper_tag_class.options.end_tag = alias['end_tag']
        #
        #     new_doc = cls.gen_doc(alias_to=cls.options.start_tag)
        #     library.tag(alias['start_tag'], init_outer(new_wrapper_tag_class, doc=new_doc))


        return cls

    return wrap



from __future__ import unicode_literals, print_function
from django.core.validators import *
from django.utils.encoding import smart_text, python_2_unicode_compatible


@python_2_unicode_compatible
class AnyValidator(object):

    validators = None

    def __init__(self, *validators):
        self.validators = validators

    def __call__(self, value):
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError:
                continue
            else:
                return

    def __str__(self):
        return 'Any validator: {}'.format(self.validators)


@python_2_unicode_compatible
class RequiresTagValidator(object):

    tags = None

    def __init__(self, *tags):
        self.tags = tags

    def __call__(self, value):
        """
        raises ValidationError.
        """

        if not value:
            return

        try:
            tag = self.get_tag(value)
        except ValueError:
            raise ValidationError('value {} not a tag.'.format(value))

        if self.tags and tag not in self.tags:
            raise ValidationError('tag {} not in allowed list: {}.'.format(tag, self.tags))

    def get_tag(self, value):
        """
        Returns tag if value has one
        :param value:
        :return:
        """
        tag = getattr(value, 'tag', None)

        if tag is None:
            raise ValueError('value {} not tag'.format(smart_text(value)))

        return tag

    def __str__(self):
        if len(self.tags) > 1:
            f = 'Required tag in: {}'
        else:
            f = 'Required tag: {}'
        return f.format(self.tags)


@python_2_unicode_compatible
class StringValidator(object):

    def __call__(self, value):
        if not isinstance(value, six.string_types):
            raise ValidationError('value is not string')

    def __str__(self):
        return 'value must be string'



from __future__ import absolute_import, print_function, unicode_literals

from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class RenderedTag(object):
    """
    Rendered tag curried string object with data, so we can interact with the data on upper levels
    """
    content = None
    tag = None
    data = None

    def __init__(self, rendered_content, tag_name, **data):
        self.content = rendered_content
        self.tag = tag_name
        self.data = data

    def __str__(self):
        """
        Return rendered content
        :return:
        """
        return self.content

    def __getattr__(self, attr):
        if attr in self.__dict__:
            return self.__dict__[attr]
        return self.data[attr]



from __future__ import absolute_import, print_function, unicode_literals

import re

from django.core.exceptions import ValidationError
from django.template import TemplateSyntaxError, Context
from django.utils.functional import cached_property
from django.utils.translation import ugettext_lazy as _

from wrapper_tag import utils
from wrapper_tag import docgen

# regex type for isinstance
REGEX_TYPE = type(re.compile(''))


class Argument(utils.TemplateMixin):

    # Tracks each time a Field instance is created. Used to retain order.
    creation_counter = 0

    # name is provided by tag metaclass
    name = None

    # logger instance
    logger = None

    _choices = None
    _default = None

    data_callback = None
    help_text = None
    readonly = False
    _tag_render_method = None
    validators = None
    extra_data = None

    # default doc_group
    doc_group = docgen.ArgumentsGroup(_('Arguments'), help_text=_('Group of generic arguments'))

    def __init__(self, help_text=None, data_callback=None, readonly=False, default=None,
                 validators=None, extra_data=None, choices=None, tag_render_method=None, **kwargs):
        # Increase the creation counter, and save our local copy.
        self.creation_counter = Argument.creation_counter
        Argument.creation_counter += 1

        self.data_callback = self._get_data_callback(data_callback)
        self.extra_data = extra_data or {}
        self.help_text = help_text
        self.readonly = readonly
        self._tag_render_method = tag_render_method
        self._default = default

        if validators and not utils.is_seq(validators):
            validators = [validators]

        self.validators = validators or []
        self._choices = choices

        super(Argument, self).__init__(**kwargs)

    def _get_data_callback(self, data_callback):
        """
        Returns data_callback from value

        @TODO: this does not work as string since we need to set data_callback from upper class (tag)
        :return:
        """

        if not data_callback:
            return None

        return data_callback

    def __repr__(self):
        """
        Representation of argument
        :return:
        """
        return '<argument:{}:{} at 0x{:x}>'.format(self.name, self.__class__.__name__, id(self))

    @property
    def tag_clean_method(self):
        return 'clean_{}'.format(self.name)

    @property
    def tag_render_method(self):
        if self._tag_render_method:
            return self._tag_render_method
        return 'render_{}'.format(self.name)

    @property
    def template_errors(self):
        return {
            'no_template': 'Argument `{}` doesn\'t provide template or template_name'.format(self.name),
        }

    @property
    def default(self):
        if callable(self._default):
            return self._default()
        return self._default

    @default.setter
    def default(self, value):
        self._default = value

    @property
    def choices(self):
        if callable(self._choices):
            return self._choices()
        return self._choices

    @choices.setter
    def choices(self, value):
        self._choices = value

    def contribute_to_class(self, tag_cls, name):
        """
        Contribute to class
        :param tag_cls:
        :param name:
        :return:
        """

        def _clean_argument_(tag, argument, value):
            """
            Dummy clean_<argument> method
            """
            return value

        # check if clean_<argument> method is defined on tag, if not create dummy one
        tcl = getattr(tag_cls, self.tag_clean_method, None)
        if tcl is None:
            setattr(tag_cls, self.tag_clean_method, _clean_argument_)

        # check if render_<argument> method is defined on tag, if not assign `render`
        trm = getattr(tag_cls, self.tag_render_method, None)
        if trm is None:
            setattr(tag_cls, self.tag_render_method, self.render)
        else:
            # verify data_callback signature
            if utils.is_template_debug():
                utils.verify_func_signature(trm, 'argument', 'data', 'context', exact_names=True,
                                            prefix="{}.{}, ".format(tag_cls.__name__, self.name))

        if callable(self.data_callback):

            # verify data_callback signature
            if utils.is_template_debug():
                utils.verify_func_signature(self.data_callback, 'tag', 'data', exact_names=True)

            tag_cls.add_data_callback(self.data_callback)

    def full_clean(self, tag, value):
        """
        Internal!

        Do not override this method, rather provide `clean` method
        """
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError:
                value = self.default
                break

        value = getattr(tag, self.tag_clean_method)(self, value)
        value = self.clean(tag, value)
        return value

    def get_tag_value(self, args, kwargs):
        """
        Get keyword value from args/kwargs
        :param args:
        :param kwargs:
        :return:
        """
        if self.readonly:
            return self.default
        return kwargs.pop(self.name, self.default)

    def clean(self, tag, value):
        """
        This method should be overridden
        :param value:
        :return:
        """
        return value

    def full_render(self, tag, data, context):
        """
        Internal!

        :param data:
        :return:
        """
        return getattr(tag, self.tag_render_method)(self, data, context)

    def render(self, argument, data, context):
        """
        Rendering function
        :param data:
        :return:
        """
        template = self.template
        if not template:
            return utils.NULL

        return template.render(Context(data))

    def gen_doc(self):
        """
        Generate documentation for argument
        :return:
        """

        doc = '``{}``'.format(self.name)

        if self.help_text:
            doc = '{} - {}'.format(doc, self.help_text)

        return doc


class Keyword(Argument):
    """
    Single keyword argument such as {% tag title="title" %}
    """

    # default doc_group
    doc_group = docgen.ArgumentsGroup(_('Keyword arguments'), priority=1000)

    def get_tag_value(self, args, kwargs):
        """
        Get keyword value from args/kwargs
        :param args:
        :param kwargs:
        :return:
        """
        value = super(Keyword, self).get_tag_value(args, kwargs)
        if self.choices and value not in self.choices:
            value = self.default
        return value

    def gen_doc(self):
        doc = super(Keyword, self).gen_doc()
        if self.choices:
            doc = '{}\n    * choices - {}'.format(doc, self.choices)
        if self.default:
            doc = '{}\n    * default - {}'.format(doc, self.default)
        return doc


class KeywordGroup(Argument):
    """
    Group of multiple keyword arguments.
    """

    source = None
    choices = None

    # default doc_group
    doc_group = docgen.ArgumentsGroup(_('Keyword arguments groups'),
                                      help_text=_('Group of keywords identified by "source". Source is a list of '
                                                  'regular expressions that keywords names match.'))

    def __init__(self, source=None, choices=None, **kwargs):
        kwargs['default'] = kwargs.pop('default', {})
        kwargs['choices'] = kwargs.pop('choices', {})
        super(KeywordGroup, self).__init__(**kwargs)

        # clean source
        self._set_source(source)

        self.choices = choices

        # check if sources are set
        if not self.readonly and not self.source and utils.get_config().template_debug:
            raise TemplateSyntaxError('keyword group must have source set.')

    def _set_source(self, source):
        """
        Clean source and convert to regular expression

        @TODO: probably source must be just simple list of strings (no flags) so user can add custom source.

        :return:
        """
        self.source = []
        if source is None:
            return

        if not utils.is_seq(source):
            source = [source]

        for item in source:
            if isinstance(item, REGEX_TYPE):
                self.source.append((item.pattern, item.flags))
            else:
                pattern = '^{}$'.format(item.replace('+', '.+').replace('*', '.*'))
                self.source.append((pattern, 0))

    @property
    def compiled_source(self):
        """
        Return list of regexes of sources.
        :return:
        """
        for item in self.source:
            yield re.compile(item[0], item[1])

    def is_source(self, source):
        """
        Check if given source is one of sources
        :param source:
        :return:
        """
        for cs in self.compiled_source:
            if cs.match(source):
                return True
        return False

    def get_tag_value(self, args, kwargs):
        """
        Get keyword value from args/kwargs
        :param args:
        :param kwargs:
        :return:
        """
        result = self.default or {}

        if self.readonly:
            return result

        for key in kwargs.keys():
            for source in self.compiled_source:
                if source.match(key):
                    result[key] = kwargs.pop(key)

        return result

    def filter_data(self, data):
        """
        Filter data for sources
        :param data:
        :return:
        """
        result = {}
        cs = self.compiled_source
        for key in data.keys():
            for source in cs:
                if source.match(key):
                    result[key] = data[key]
        return result

    def gen_doc(self):
        doc = super(KeywordGroup, self).gen_doc()
        source = ['"{}"'.format(x[0]) for x in self.source]
        if source:
            doc = '{}\n    * source - [{}]'.format(doc, ", ".join(source))
        if self.choices:
            doc = '{}\n    * choices - {}'.format(doc, self.choices)
        if self.default:
            doc = '{}\n    * default - {}'.format(doc, self.default)
        return doc


class Event(Argument):
    """
    Event is shorthand for event arguments.

    Event has additional functionality to add `events` on RenderedTag.
    """

    # default doc_group
    doc_group = docgen.ArgumentsGroup(_('Events'), help_text=_('Event handlers that are available on rendered '
                                                               'tag `events` dictionary'), priority=0)

    def contribute_to_class(self, tag_cls, name):
        """
        Patch render_wrapper_tag method on tag, to set 'events' on RenderedTag
        :param tag_cls: tag class
        :param name:
        :return:
        """
        super(Event, self).contribute_to_class(tag_cls, name)

        if getattr(tag_cls, '__events_patched', False):
            return

        def custom_render(orig):
            def render(tag, tag_kwargs, context):
                return orig(tag, tag_kwargs, context)
            return render
        tag_cls.render_tag = custom_render(tag_cls.render_tag)

        # mark tag as patched
        setattr(tag_cls, '__events_patched', True)


class Method(Argument):

    # default doc_group
    doc_group = docgen.ArgumentsGroup(_('Methods'), help_text=_('Methods that are available on rendered '
                                                                'tag `methods` dictionary'), priority=0)



from __future__ import absolute_import, print_function, unicode_literals

import copy
from collections import OrderedDict

from django.core.exceptions import ImproperlyConfigured
from django.template.exceptions import TemplateSyntaxError
from django.template import Node
from django.utils import six

from wrapper_tag import utils
from wrapper_tag import arguments
from wrapper_tag import rendered

from logging import getLogger


logger = getLogger('wrapper_tag.tags')


class TagOptions(utils.TemplateMixin):
    """
    Tag options stored as `Meta` class on tag
    """

    _start_tag = None
    end_tag = None

    def __init__(self, options=None, tag_name=None):
        self.start_tag = getattr(options, 'start_tag', None)

        # if no start_tag is given, we construct it from tag class name
        if not self.start_tag:
            if tag_name.lower().endswith('tag'):
                tag_name = tag_name[:-3]
            self.start_tag = utils.camel_to_separated(tag_name)

        end_tag = getattr(options, 'end_tag', None)
        if end_tag:
            self.end_tag = end_tag

        self.as_var_only = bool(getattr(options, 'as_var_only', False))
        self.aliases = getattr(options, 'aliases', [])
        self.tag_name = tag_name

        super(TagOptions, self).__init__(template=getattr(options, 'template', None),
                                         template_name=getattr(options, 'template_name', None))

    @property
    def start_tag(self):
        return self._start_tag

    @start_tag.setter
    def start_tag(self, start_tag):
        self._start_tag = start_tag
        self.end_tag = 'end:{}'.format(start_tag)

    @property
    def template_errors(self):
        return {
            'no_template': 'Tag `{}` doesn\'t provide template or template_name'.format(self.start_tag),
        }

    def __repr__(self):
        return '<Options for %s>' % self.tag_name


# noinspection PyUnresolvedReferences
class TagMetaclass(type):
    """
    Metaclass that collects Arguments declared on the base classes. It also prepares tag options.
    """

    def __new__(mcs, name, bases, attrs):
        options = TagOptions(attrs.get('Meta', None), tag_name=name)
        current_arguments = []

        # add logger if not provided
        if 'logger' not in attrs:
            attrs['logger'] = utils.get_sub_logger(logger, options.start_tag)

        for key, value in list(attrs.items()):
            if isinstance(value, arguments.Argument):
                # set argument name
                value.name = key
                current_arguments.append((key, value))
                attrs.pop(key)

        current_arguments.sort(key=lambda x: x[1].creation_counter)

        attrs['_declared_arguments'] = OrderedDict(current_arguments)

        # # set _decorated_function for template Library.tag so it gets correct name for tag.
        # def fake_func():
        #     pass  # pragma: no cover
        # fake_func.__name__ = options.start_tag
        # attrs['_decorated_function'] = fake_func
        attrs['name'] = options.start_tag

        # create new class
        new_class = super(TagMetaclass, mcs).__new__(mcs, name, bases, attrs)
        new_class.options = options

        # prepare declared arguments
        declared_arguments = OrderedDict()
        for base in reversed(new_class.__mro__):
            if hasattr(base, '_declared_arguments'):  # inherited tag
                declared_arguments.update(base._declared_arguments)
            else:  # tag mixin support
                base_arguments = []
                for key, value in list(base.__dict__.items()):
                    if isinstance(value, arguments.Argument):
                        value.name = key
                        base_arguments.append((key, value))

                # sort arguments
                base_arguments.sort(key=lambda x: x[1].creation_counter)
                declared_arguments.update(OrderedDict(base_arguments))

        # store declared_arguments
        new_class._declared_arguments = declared_arguments

        return new_class

    def __init__(cls, name, bases, attrs):
        """
        __init__ is called at class creation time
        :param name: class name
        :param bases: base classes
        :param attrs: attributes dict
        """
        # Add arguments to tag class
        cls.arguments = copy.deepcopy(cls._declared_arguments)

        super(TagMetaclass, cls).__init__(name, bases, attrs)

        # contribute argument to tag class
        for name, argument in six.iteritems(cls.arguments):
            argument.logger = utils.get_sub_logger(cls.logger, argument.name)
            argument.contribute_to_class(cls, name)


# noinspection PyUnresolvedReferences
class BaseTag(object):
    """
    BaseTag provides compile function (constructor) and also render method
    """
    args = None
    kwargs = None
    varname = None
    nodelist = None
    data_callbacks = []
    arguments = []

    def __init__(self, parser, token):
        """
        Compile function
        :param parser:
        :param token:
        """

        # empty data_callbacks
        self.data_callbacks = []

        # parse here all args, kwargs...
        self.nodelist = parser.parse((self.options.end_tag,))

        bits = token.split_contents()[1:]

        # parse "as" variable
        if len(bits) >= 2 and bits[-2] == 'as':
            self.varname = bits[-1]
            bits = bits[:-2]

        # parse args and kwargs
        self.args, self.kwargs = utils.parse_bits(parser, bits, self.options.start_tag)

        # die token die
        parser.delete_first_token()

    @classmethod
    def add_data_callback(cls, data_callback):
        """
        Adds single data_callback function to tag class
        :param data_callback:
        :return:
        """
        cls.data_callbacks.append(data_callback)

    def get_tag_data(self, context):
        """
        Return tag kwargs for render
        :param context:
        :return:
        """
        kwargs = dict([(key, value.resolve(context)) for key, value in self.kwargs.items()])
        args = list([value.resolve(context) for value in self.args])
        tag_data = {}

        # iterate over arguments and call get_tag_value on it
        for _, argument in six.iteritems(self.arguments):
            value = argument.get_tag_value(args, kwargs)

            # clean the value
            value = argument.full_clean(self, value)

            if value is not None:
                tag_data[argument.name] = value

        # unhandled args
        if utils.get_config().template_debug and args:
            raise TemplateSyntaxError('Tag `{}` received unhandled args: {}'.format(self.options.start_tag, args))

        # unhandled kwargs
        if utils.get_config().template_debug and kwargs:
            raise TemplateSyntaxError('Tag `{}` received unhandled kwargs: {}'.format(
                self.options.start_tag, kwargs.keys()))

        # update data
        for data_callback in self.data_callbacks:
            tag_data = data_callback(self, tag_data)

        # render all arguments
        for _, argument in six.iteritems(self.arguments):
            argument.logger.debug('Rendering with: %s', tag_data)

            tmp = argument.full_render(self, tag_data, context=context)
            if tmp is utils.NULL or tmp is None:
                continue
            tag_data['{}__rendered'.format(argument.name)] = tmp

        return tag_data

    def render(self, context):
        """
        Render method does multiple steps

        * resolve args/kwargs
        * iterate over all arguments and call get_tag_value
        """
        tag_kwargs = self.get_tag_data(context)
        self.logger.debug("Rendering with: %s", tag_kwargs)

        # render content which is isolated from tag data
        tag_kwargs['content'] = self.nodelist.render(context)

        return self.render_tag(tag_kwargs, context)

    def render_tag(self, tag_kwargs, context):
        """
        The method you should override in your custom tags
        """
        with context.push(**tag_kwargs):
            template = self.options.template
            if template is None:
                raise ImproperlyConfigured("tag: {}, error: please provide meta.template_name or meta.template".format(
                                           self.options.start_tag))
            tmp = template.render(context)
        return rendered.RenderedTag(tmp, self.options.start_tag, **tag_kwargs)


class Tag(six.with_metaclass(TagMetaclass, BaseTag, Node)):
    """
    Main tag to inherit from
    """



from __future__ import absolute_import, print_function, unicode_literals
from django import template
from django.utils.safestring import mark_safe

import wrapper_tag
from wrapper_tag import tag
from wrapper_tag import arguments

register = template.Library()


@wrapper_tag.register_tag(register)
class TestTag(tag.Tag):
    title = arguments.Keyword()
    name = arguments.Keyword()
    on_click = arguments.Event()

    class Meta:
        template = "<test{{ title__rendered }}>{{ content }}</test>"

    def render_title(self, argument, data, context):
        if argument.name not in data:
            return None
        return mark_safe(' title="{}"'.format(data[argument.name]))

    def render_name(self, argument, data, context):
        if argument.name not in data:
            return None
        return mark_safe(' name="{}"'.format(data[argument.name]))

    def render_on_click(self, argument, data, context):
        return "on_click"



from django.apps import AppConfig
from django.conf import settings
from logging import getLogger

try:
    # Django versions >= 1.9
    from django.utils.module_loading import import_module
except ImportError:
    # Django versions < 1.9
    from django.utils.importlib import import_module

import unipath


from wrapper_tag import utils


class WrapperTagConfig(AppConfig):
    name = 'wrapper_tag'
    verbose_name = 'wrapper tags'

    logger = getLogger('wrapper_tag')
    template_debug = utils.is_template_debug()

    def get_templatetag_libs(self):
        """
        Returns all templatetag libraries
        :return:
        """
        result = {}
        for installed_app in getattr(settings, 'INSTALLED_APPS', []):

            tt = '{}.templatetags'.format(installed_app)

            try:
                module = import_module(tt)
            except ImportError:
                continue

            for pf in unipath.Path(module.__file__).parent.listdir():
                if not (pf.isfile() and pf.ext.lower() == '.py' and not pf.name.startswith('_')):
                    continue
                ttfull = '{}.{}'.format(tt, pf.stem)

                try:
                    imported = import_module(ttfull)

                    register = getattr(imported, 'register', None)
                    if register:
                        result[str(pf.stem)] = register

                except ImportError:
                    continue

        return result



from __future__ import absolute_import, print_function, unicode_literals

import random
import re

import six
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from wrapper_tag import arguments, validators, utils


IDENTITY_ID_RAND_MAX = 2 ** 32


class TagAttributes(object):

    attrs = arguments.KeywordGroup(readonly=True)

    def render_attrs(self, argument, data, context):
        attrs = data.get(argument.name, {})
        full = ' '.join(['{}="{}"'.format(key, value) for key, value in six.iteritems(attrs)])
        full = ' ' + full if full else ''
        return mark_safe(full)


def id_data_callback(tag, data):
    """
    Add id to attrs
    """
    if 'attrs' in data:
        data['attrs']['id'] = data['id']
    return data


class Identity(TagAttributes):

    id = arguments.Keyword(data_callback=id_data_callback)
    name = arguments.Keyword()

    def clean_id(self, argument, value):
        """
        Try to clean id, if not given generate one
        """
        if not value:
            return 'id_{}'.format(random.randint(0, IDENTITY_ID_RAND_MAX))
        return value

    def render_id(self, argument, data, context):
        """
        Custom render for argument `id`
        """
        if 'id' in data and data['id']:
            return format_html('id="{id}"', **data)

    def render_name(self, argument, data, context):
        """
        Custom rende method for argument `name`
        """
        if 'name' in data and data['name']:
            return format_html('name="{name}"', **data)


def css_class_data_callback(tag, data):
    """
    Check if TagAttributes
    """
    if 'attrs' in data and 'css_class' in data:
        data['attrs']['class'] = data['css_class']
    return data


class CssClass(object):
    """
    Adds ability to set cusom css_class
    """
    css_class = arguments.Keyword(default='', help_text='Additional css class (e.g. ``"class1 class2"``)',
                                  data_callback=css_class_data_callback)


"""
Old mixins




"""


class JsEvents(object):

    js_events = arguments.KeywordGroup(readonly=True)

    def render_js_events(self, argument, data, context):
        """
        Renders javascript event handlers.
        :param argument:
        :param data:
        :param context:
        :return:
        """
        if 'id' not in data or not data['id']:
            return ''

        js_events = data.get('js_events', {})
        OBJ_NAME = '$obj'

        lines = []

        for js_event in js_events.keys():
            if js_event not in argument.extra_data:
                continue

            lines.append(
                argument.extra_data[js_event].format(object=OBJ_NAME, function=js_events[js_event])
            )

        if not lines:
            return ''

        lines.insert(0, 'var {obj_name} = $("#{id}");'.format(obj_name=OBJ_NAME, **data))
        result = format_html('<script type="text/javascript">$(function() {{ {script} }});</script>',
                             script=mark_safe("".join(lines)))
        return result


class Data(object):
    """
    Data handles kwarguments for `data-` html attributes.
    """
    data = arguments.KeywordGroup('data_*', help_text='data attributes, data_* will be converted to data-*')

    def render_data(self, argument, data, context=None):
        """
        Render data-* html attributes
        :param attr: attribute instance
        :param data: whole tag data
        :param context: context for rendering
        :return:
        """
        value = data.get(argument.name, {})
        return ' '.join(['{}="{}"'.format(k.replace('_', '-'), v) for k, v in six.iteritems(value)])


class Tag(object):
    """
    Add ability to set custom tag.
    """

    # override these properties to customize tag mixin
    TAG_CHOICES = ['div']
    TAG_DEFAULT = 'div'

    tag = arguments.Keyword(help_text='tag', choices=lambda: Tag.TAG_CHOICES, default=lambda: Tag.TAG_DEFAULT,
                            tag_render_method="render_html_tag")


class Hyperlink(object):
    """
    Adds ability to provide href functionality
    """
    href = arguments.Keyword(help_text=_('raw href (url takes precedence)'))
    url = arguments.KeywordGroup(('url', 'url_kwarg_*', 'url_arg_*'),
                                 help_text=_('if provided, href will be set with call to url reverse'))
    url_query = arguments.Keyword(help_text=_('if provided, will be added as url query params.'))

    def render_href(self, argument, data, context=None):
        """
        custom href rendering
        :param attribute:
        :param data:
        :return:
        """
        if 'url' in data and 'reversed' in data['url']:
            return format_html('href="{reversed}"'.format(**data['url']))

        if 'href' in data and data['href']:
            return format_html('href="{href}"'.format(**data))

    def clean_url(self, argument, data):
        """
        clean_url tries to resolve given url

        @TODO: fix this
        :param argument:
        :param data:
        :return:
        """

        if 'url' not in data:
            return data

        kwarg_keys = utils.find_elements('url_kwarg_*', data.keys())
        arg_keys = utils.find_elements('url_arg_*', sorted(data.keys()))
        url_kwargs = {k.lstrip('url_kwarg_'): v for k, v in six.iteritems(data) if k in kwarg_keys}
        url_args = [data[v] for v in arg_keys]

        data['reversed'] = reverse(data['url'], args=url_args, kwargs=url_kwargs)
        if 'url_query' in data and data['url_query']:
            data['reversed'] = '{}?{}'.format(data['reversed'], data['url_query'])
        return data


class Icon(object):

    icon = arguments.Keyword(default='', help_text=_('fontawesome icon name (e.g. `user`)'))

    def render_icon(self, argument, data, context=None):
        # hack for now
        # data['iconset'] = 'fa'
        # print '<i class="{iconset} {iconset}-{name}"></i>'.format(**data)
        # @TODO: check this code
        # self.logger.debug("this is icon data %s", data)

        if 'icon' in data and data['icon']:
            return format_html('<i class="fa fa-{}"></i>', data['icon'])


class Size(object):

    size = arguments.KeywordGroup(('xs', 'sm', 'md', 'lg'))

    def render_size(self, argument, data, context):
        lines = []
        for key, value in six.iteritems(argument.filter_data(data.get(argument.name, {}))):
            lines.append(self.format_size(key, value))
        return " ".join(lines)

    def format_size(self, key, value):
        return format_html('col-{key}-{value}'.format(key=key, value=value))


def tooltip_contribute_data(tag, data):
    """
    @TODO: fix this
    :param tag:
    :param data:
    :return:
    """
    return data

    value = data['tag']

    if value is None:
        return

    from wrapper_tag import RenderedTag

    if 'data' not in data:
        data['data'] = {}

    if isinstance(value, RenderedTag):
        data['data']['rel'] = 'tooltip'
        data['data']['data_placement'] = value.attributes['position']
        data['data']['original_title'] = value.attributes['content']
        data['data']['title'] = value.attributes['content']
    else:
        data['data']['rel'] = 'tooltip'
        data['data']['original_title'] = value
        data['data']['title'] = value

    return data


class Tooltip(Data):
    tooltip = arguments.Keyword(help_text='Tooltip support (either rendered tooltip or string',
                                validators=validators.AnyValidator(
                                    validators.RequiresTagValidator('tooltip'),
                                    validators.StringValidator()
                                ),
                                data_callback=tooltip_contribute_data
                                )


class Common(Identity, CssClass, Data, Tag):
    """
    common mixin used in most of wrapper tags.
    """


class ColorMixin(object):

    AVAILABLE_COLORS = (
        'green', 'greenDark', 'greenLight',
        'purple', 'magenta',
        'pink', 'pinkDark',
        'blue', 'blueLight', 'blueDark',
        'teal',
        'yellow',
        'orange',
        'orangeDark',
        'red',
        'redLight',
    )

    def get_bg_color_css_class(self, value):
        if not value:
            return
        return 'bg-color-{}'.format(value)

    def get_text_color_css_class(self, value):
        if not value:
            return
        return 'txt-color-{}'.format(value)


class BackgroundColorMixin(ColorMixin):

    bg_color = arguments.Keyword(choices=ColorMixin.AVAILABLE_COLORS)

    def render_bg_color(self, argument, data, context):
        value = data.get(argument.name, None)
        if not value:
            return value
        return self.get_bg_color_css_class(value)


class TextColorMixin(ColorMixin):
    AVAILABLE_COLORS = ColorMixin.AVAILABLE_COLORS + (
        'muted',
        'primary',
        'success',
        'danger',
        'warning',
        'info',
    )

    text_color = arguments.Keyword(choices=AVAILABLE_COLORS)

    def render_text_color(self, argument, data, context=None):
        value = data.get(argument.name, None)
        if not value:
            return
        return self.get_text_color_css_class(value)


class ModalJsEvents(JsEvents):
    js_events = arguments.KeywordGroup(('on_show', 'on_after_show', 'on_hide', 'on_after_hide', 'on_data_load'),
                                  extra_data={
                                      'on_show': '{object}.on("show.bs.modal", {function});',
                                      'on_after_show': '{object}.on("shown.bs.modal", {function});',
                                      'on_hide': '{object}.on("hide.bs.modal", {function});',
                                      'on_after_hide': '{object}.on("hidden.bs.modal", {function});',
                                      'on_data_load': '{object}.on("loaded.bs.modal", {function});',
                                  })
